from .model import Model
from .simulator import Simulator
from .report import Report
from .enumtypes import DictionaryType, LossType
from .gspread_client import GspreadClient
from .gdrive_client import GdriveClient
from .utils import *
from .demo import *

## Report Service Folder Structure
images_path = report_service_conf['images_path']
reports_path = report_service_conf['reports_path']

try:
	os.makedirs(images_path)
	print(f"Folder created at '{images_path}'.")
except OSError as e:
	print(f"Images Folder at '{images_path}'.")

try:
	os.makedirs(reports_path)
	print(f"Folder created at '{reports_path}'.")
except OSError as e:
	print(f"Reports Folder at '{reports_path}'.\n")



import os
import itertools
import googleapiclient
import numpy as np
import matplotlib.pyplot as plt

from IPython.display import clear_output
from shapely.geometry import LineString
from tabulate import tabulate

from matplotlib.backends.backend_pdf import PdfPages
from reportlab.lib.pagesizes import letter
from reportlab.platypus import SimpleDocTemplate, Paragraph, Spacer

from .enumtypes import LossType
from .utils import cls, report_service_conf
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .simulator import Simulator


class Report:

  """Report of the executed simulations """

  def __init__(self, sim):
    """
    Args:
      sim (Simulator): Simulator object

    Attributes:
      sim (Simulator): Simulator object
      iter_N (dict): Number of iterations for each dimension and loss type
      max_iter_N (list): Maximum number of iterations for each dimension
      loss_N (dict): Loss for each dimension and loss type
      loss_bayes (dict): Bayes loss for each dimension
      d (dict): Dictionary size for each dimension
      duration (float): Duration of the simulation
      time_spent (dict): Time spent for each dimension and loss type
      sim_tag (dict): Simulator object attributes
      model_tag (dict): Model object attributes
      loss_plot (matplotlib.figure.Figure): Loss plot

    Raises:
      TypeError: If sim is not a Simulator object


    """

    if not isinstance(sim, Simulator):
      raise TypeError("sim must be a Simulator object")

    self.sim = sim
    self.iter_N = {dim: {loss_type: [] for loss_type in sim.loss_types} for dim in sim.dims}
    self.max_iter_N = []
    self.loss_N = {dim: {loss_type: [] for loss_type in sim.loss_types} for dim in sim.dims}
    self.loss_bayes = {dim : 0 for dim in sim.dims}
    self.d = {dim: 0 for dim in sim.dims}
    self.duration = 0
    self.time_spent = {loss_type: 0.0 for loss_type in sim.loss_types}
    self.time_spent.update({'n': [0.0 for n in self.sim.model.N]})
    self.time_spent.update({d:0.0 for d in self.sim.dims})
    self.time_spent.update({'total':0.0})
    self.sim_tag = dict(itertools.islice(self.sim.__dict__.items(), 7))
    self.model_tag = dict(itertools.islice(self.sim.model.__dict__.items(), 6))
    self.loss_plot = None

  def compile_delta_L_(self):
    """return :math:`∆L` estimations


    Stochastic error:
      - :math:`∆L_1 = L(\hat{h}(D)) − min_{h∈H} L(h)`

    Estimation error of :math:`L(\hat{h}(D))`:
      - :math:`∆L_2 = |L(\hat{h}(D)) − \hat{L}(\hat{h}(D))|`

    :param self: report object
    :type self: Report
    :return: delta_L_ = (delta_L1, delta_L2)
    :rtype: tuple of dicts

    """

    loss_N = self.loss_N
    dims = self.sim.dims
    loss_bayes = self.loss_bayes
    N = self.sim.model.N

    dims_aux = []
    for d in dims:
      if loss_bayes[d]:
        dims_aux.append(d)

    delta_L1 = {dim: [loss_N[dim][LossType.THEORETICAL.value][i] - loss_bayes[dim] if loss_bayes[dim] > 0 else 0 for i in range(len(N))] for dim in dims }  if LossType.THEORETICAL.value in self.sim.loss_types else []

    delta_L2 = {dim: [abs(loss_N[dim][LossType.THEORETICAL.value][i] - loss_N[dim][LossType.EMPIRICALTRAIN.value][i]) for i in range(len(N))]  for dim in dims} if LossType.EMPIRICALTRAIN.value in self.sim.loss_types else []

    delta_Ltest = {dim: np.mean(loss_N[dim][LossType.THEORETICAL.value]) - loss_bayes[dim] if loss_bayes[dim] > 0 else 0 for dim in dims}

    delta_L_ = (delta_L1, delta_L2)
    return delta_L_

  def intersection_point_(self, dims, loss_type):
    """return intersection points between Loss curves of a pair of compared dimensionalyties

    :param self: report object
    :type self: Report
    :param dims: a pair of dimensionalyties to be compared
    :type dims: tuple of int or list of int
    :param: loss estimation method
    :type loss_type: str

    :return: intersection points between Loss curves of a pair of compared dimensionalyties
    :rtype: list of lists

    """


    ydata1 = self.loss_N[dims[0]][loss_type]
    ydata2 = self.loss_N[dims[1]][loss_type]
    xdata = self.sim.model.N[:len(ydata1)]

    line_1 = LineString(np.column_stack((np.log2(xdata), ydata1)))
    line_2 = LineString(np.column_stack((np.log2(xdata), ydata2)))
    intersection = line_1.intersection(line_2)

    intersection_points = []
    n_star = []

    if str(intersection)[0:10] == 'MULTIPOINT':
      for i in range(0, len(intersection.geoms)):
        if(intersection.geoms[-(i+1)].x > 1):
          intersection_points.append([intersection.geoms[-(i+1)].x,intersection.geoms[-(i+1)].y])
          n_star.append(2**intersection.geoms[-(i+1)].x)

    elif str(intersection)[0:10] == 'GEOMETRYCO' or str(intersection)[0:10] == 'MULTILINES' or str(intersection)[0:10] == 'LINESTRING':
      n_star.append('N/A')

    elif str(intersection) == 'LINESTRING EMPTY'  or  str(intersection) == 'LINESTRING Z EMPTY' :
      pass
    else:
      if(intersection.x > 1):
        n_star.append(2**intersection.x)
        intersection_points.append([intersection.x,intersection.y])

    return intersection_points, n_star

  def compile_compare(self, dims=None):
    """return compare_report images for pair of compared dimensionalyties

    Parameters:
      dims (tuple of int or list of int): a pair of dimensionalyties to be compared

    Returns:
      tuple[dict, dict, dict[int | str, str], dict[int | str, Any], dict, dict[str, Any], dict[str, Any]]: compare_report images for pair of compared dimensionalyties

    """

    dims = dims if dims else self.sim.dims[-2:]

    intersection_point_ = {loss_type : self.intersection_point_( dims, loss_type)[0] for loss_type in self.sim.loss_types}
    n_star_ = {loss_type : self.intersection_point_( dims, loss_type)[1] for loss_type in self.sim.loss_types}

    loss_N = {d : { loss_type : self.loss_N[d][loss_type] for loss_type in list(self.loss_N[d].keys())} for d in dims}
    iter_N = {d : { loss_type : self.iter_N[d][loss_type] for loss_type in list(self.iter_N[d].keys())} for d in dims}

    intersection_point_dict = { loss_type: { 'log_2(N*)' : np.array(intersection_point_[loss_type]).T[0] if intersection_point_[loss_type] else ['n/a'],
                                                 'P(E)': np.array(intersection_point_[loss_type]).T[1] if intersection_point_[loss_type] else ['n/a'],
                                                 'N*' : n_star_[loss_type] if n_star_[loss_type] else ['NO INTERSECT'] }
                               for loss_type in self.sim.loss_types}

    bayes_ratio = self.loss_bayes[dims[0]]/self.loss_bayes[dims[1]] if  self.loss_bayes[dims[1]] > 0 else 'n/a'
    bayes_diff = self.loss_bayes[dims[0]] - self.loss_bayes[dims[1]] if  self.loss_bayes[dims[1]] > 0 else 'n/a'

    loss_bayes = {dims[0] : self.loss_bayes[dims[0]] , dims[1] : self.loss_bayes[dims[1]] }

    loss_bayes.update({'ratio': bayes_ratio , 'diff': bayes_diff })

    d_ratio = self.d[dims[0]]/self.d[dims[1]]
    d_diff = self.d[dims[0]] - self.d[dims[1]]

    d = {dims[0] : self.d[dims[0]] , dims[1] : self.d[dims[1]]}

    d.update({'ratio': d_ratio , 'diff': d_diff })

    compare = (loss_N, iter_N, loss_bayes, d, intersection_point_dict, self.model_tag, self.sim_tag)
    return compare

  def compile_N(self, dims=(2,3)):
    """return N* images for report compilation. N* is a threshold beyond which the presence of a new feature X_d becomes advantageous, if the other features [X_0...X_d-1] are already present.

    :self: report object
    :type self: Report
    :param dims: a pair of dimensionalyties to be compared
    :type dims: tuple of int or list of int
    :return: N* images for report compilation
    :rtype: dict

    """

    intersection_point_t, n_star_t = self.intersection_point_( dims, 'THEORETICAL')
    intersection_point_e, n_star_e = self.intersection_point_( dims, 'EMPIRICAL_TEST')

    log2_N_star_dict = {'THEORETICAL': np.array(intersection_point_t[-1]).T[0] if intersection_point_t else 0,
                        'EMPIRICAL_TEST': np.array(intersection_point_e[-1]).T[0] if intersection_point_e else 0}

    bayes_ratio = self.loss_bayes[dims[0]]/self.loss_bayes[dims[1]] if  self.loss_bayes[dims[1]] > 0 else 'n/a'
    bayes_diff = self.loss_bayes[dims[0]] - self.loss_bayes[dims[1]] if  self.loss_bayes[dims[1]] > 0 else 'n/a'

    loss_bayes = {dims[0]: self.loss_bayes[dims[0]], dims[1]: self.loss_bayes[dims[1]]}
    loss_bayes.update({'ratio': bayes_ratio, 'diff': bayes_diff})

    d_ratio = self.d[dims[0]] / self.d[dims[1]]
    d_diff = self.d[dims[0]] - self.d[dims[1]]

    d = {dims[0]: self.d[dims[0]], dims[1]: self.d[dims[1]]}
    d.update({'ratio': d_ratio, 'diff': d_diff})

    loss_types = self.sim.loss_types
    loss_N_0 = [self.loss_N[dims[0]][loss_type][i] if i<10 else self.loss_bayes[dims[0]] for loss_type in loss_types  for i in range(min(len(self.loss_N[dims[0]][loss_type])+1,11))]
    loss_N_1 = [self.loss_N[dims[1]][loss_type][i] if i<10 else self.loss_bayes[dims[1]] for loss_type in loss_types  for i in range(min(len(self.loss_N[dims[1]][loss_type])+1,11))]

    dims = self.sim.dims

    time_spent_gen = [self.duration, self.time_spent['total']]
    time_spent_loss_type = [self.time_spent[loss_type] for loss_type in loss_types]
    time_spent_n = self.time_spent['n']
    time_spent_dim = [self.time_spent[dim] for dim in dims]

    time_ratio_gen = [ 1 ,self.time_spent['total']/self.duration]
    time_ratio_loss_type = [self.time_spent[loss_type]/self.duration for loss_type in loss_types]
    time_ratio_n = [self.time_spent['n'][i]/self.duration for i in range(len(self.time_spent['n']))]
    time_ratio_dim = [self.time_spent[dim]/self.duration for dim in dims]

    iter_ratio_per_loss_type_aux = [self.iter_N[d][loss_type][i]/self.max_iter_N[i] for loss_type in loss_types for d in dims for i in range(10)]
    index = 10*len(dims)
    iter_ratio_per_loss_type = [np.average(iter_ratio_per_loss_type_aux[index*i:index*(i+1)]) for i in range(len(loss_types))]

    iter_ratio_per_n_aux = [ self.iter_N[d][loss_type][i]/self.max_iter_N[i] for i in range(10) for loss_type in loss_types for d in dims]
    index = len(loss_types)*len(dims)
    iter_ratio_per_n = [np.average(iter_ratio_per_n_aux[index*i:index*(i+1)]) for i in range(10)]

    iter_ratio_per_dim_aux = [ self.iter_N[d][loss_type][i]/self.max_iter_N[i] for d in dims for loss_type in loss_types for i in range(10)]
    index = len(loss_types)*10
    iter_ratio_per_dim = [np.average(iter_ratio_per_dim_aux[index*i:index*(i+1)]) for i in range(len(dims))]

    cost = ( time_spent_gen, time_spent_loss_type, time_spent_n, time_spent_dim, time_ratio_gen , time_ratio_loss_type, time_ratio_n, time_ratio_dim, iter_ratio_per_loss_type, iter_ratio_per_n, iter_ratio_per_dim)

    N_report_params = (self.sim.model.sigma, self.sim.model.rho, loss_bayes, d, log2_N_star_dict, loss_N_0, loss_N_1) + cost

    return N_report_params

  def save_loss_plot_as_png(self, export_path=None, verbose=True):
    """
    Save a matplotlib Figure object as a PNG image.

    Parameters:
        export_path (str): The file path where the PNG image will be saved.
        verbose (bool): If True, print the export path.
    Returns:
        None
    """
    if self.loss_plot is not None:
      if export_path is None:
        export_path = report_service_conf['images_path']
        export_path += 'loss' + str(self.sim.model.params)
        export_path += '[test]' if (self.sim.iters_per_step * self.sim.max_steps < 1000) else ''
        export_path += '.png'
      elif not export_path.endswith(".png"):
        export_path = report_service_conf['images_path']
        export_path += 'loss' + str(self.sim.model.params)
        export_path += '[test]' if (self.sim.iters_per_step * self.sim.max_steps < 1000) else ''
        export_path += '.png'

      if not os.path.exists(export_path):
        try:
          self.loss_plot.savefig(export_path, format="png", dpi=300)
          if verbose:
            print(f"Figure saved as: {export_path}")
        except Exception as e:
          print(f"Failed to save the figure: {e}")
      else:
        if verbose:
          print(f"File already exists: {export_path}")
    else:
      if verbose:
        print("No figure to save.")


  def upload_loss_plot_to_drive(self, gdc, verbose=True):
    """
    Upload a matplotlib Figure object as a PNG image to Google Drive.

    Parameters:
        gdc (GoogleDriveClient): The GoogleDriveClient object.
        verbose (bool): If True, print the export path.
    Returns:
        None
    """
    if self.loss_plot is not None:
      export_path = report_service_conf['images_path']
      export_path += 'loss' + str(self.sim.model.params)
      export_path += '[test]' if (self.sim.iters_per_step * self.sim.max_steps < 1000) else ''
      export_path += '.png'

      if not os.path.exists(export_path):
        self.save_loss_plot_as_png(export_path=export_path, verbose=verbose)

      gdrive_images_folder_path = 'slacgs.demo.' + gdc.gdrive_account_email + '/images'

      if not gdc.folder_exists_by_path(gdrive_images_folder_path):
        folder_id = gdc.create_folder('images', gdc.get_folder_id_by_name('slacgs.demo.' + gdc.gdrive_account_email), verbose=verbose)
      else:
        folder_id = gdc.get_folder_id_by_path(gdrive_images_folder_path)

      gdc.upload_file_to_drive(export_path, folder_id, verbose=verbose)
    else:
      if verbose:
        print("No figure to upload.")


  def plot_with_intersection(self):
    """plot Loss curves of a pair of compared dimensionalyties with intersection points between them

    Returns:
      matplotlib.figure.Figure: The figure object.

    """

    sim_dims = self.sim.dims

    unique_pairs = []
    n = len(sim_dims)

    for i in range(n):
      for j in range(i + 1, n):
        pair = (sim_dims[i], sim_dims[j])
        unique_pairs.append(pair)

    Xdata = np.log2(self.sim.model.N)[:len(self.loss_N[sim_dims[0]][LossType.THEORETICAL.value])]
    Y_data = self.loss_N

    columns = len(self.sim.loss_types)
    # Create the figure and three subplots
    fig, axs = plt.subplots(1, columns, figsize=(14, 4))

    for i, loss_type in enumerate(self.sim.loss_types):
      for d in sim_dims:
        axs[i].plot(Xdata, Y_data[d][loss_type], label='dim = ' + str(d))

      if len(self.loss_N[sim_dims[0]][loss_type]) > 1:
        for dims in unique_pairs:
          intersection_points, n_star = self.intersection_point_(dims, loss_type)
          if len(intersection_points) > 0:
            for j in range(0, len(intersection_points)):
              axs[i].plot(intersection_points[j][0], intersection_points[j][1], 'ro')
              axs[i].text(intersection_points[j][0], intersection_points[j][1],
                          '(' + "{:.1f}".format(intersection_points[j][0]) + ',' + "{:.2f}".format(
                            intersection_points[j][1]) + ')')

      axs[i].set_title(loss_type)
      axs[i].set_xlabel('$\log_2(n)$')
      axs[i].set_ylabel('$P(error)$')
      axs[i].set_xlim([0, max(12, len(self.sim.model.N))])
      axs[i].set_ylim([0, 1])
      axs[i].legend()

    # Show the plot
    plt.tight_layout()

    return fig

  def print_loss(self):

    for loss_type in self.sim.loss_types:
      data = np.array([ np.round(self.loss_N[dim][loss_type],4) for dim in self.sim.dims]).T.tolist()

      ## add index column
      indexed_data = [[int(2 ** (i + 1))] + sublist for i, sublist in enumerate(data)]

      ## make table and print
      table = tabulate(indexed_data, tablefmt='grid', headers=['N'] + [str(dim) + ' feat' for dim in self.sim.dims])
      print('\n',loss_type, ' Loss: ')
      print(table)

      data = np.array([self.iter_N[dim][loss_type] for dim in self.sim.dims]).T.tolist()

      ## add index column
      indexed_data = [[int(2 ** (i + 1))] + sublist for i, sublist in enumerate(data)]

      ## make table and print
      table = tabulate(indexed_data, tablefmt='grid', headers=['N'] + [str(dim) + ' feat' for dim in self.sim.dims])
      print('\n', loss_type, ' # iterations: ')
      print(table)

  def print_N_star(self, dims_to_compare=None):
    """Print N_star for theoretical and empirical loss

    Parameters:
      dims_to_compare (list of int or tuple of int): list of dimensionalities to be compared

    Returns:
      None

    Raises:
      TypeError: if dims_to_compare is not a list of int or tuple of int;
      ValueError: if the number of compared dimensionalities is not 2;
                  if the list of dioemnsionalities to be compared is not a subset of the list of simulated dimensionalities

    """

    dims_to_compare = self.sim.dims[:2] if dims_to_compare is None else dims_to_compare

    intersection_point_t, n_star_t = self.intersection_point_( dims_to_compare, 'THEORETICAL')
    intersection_point_e, n_star_e = self.intersection_point_( dims_to_compare, 'EMPIRICAL_TEST')

    data_theo = [['THEO'] + [np.round(intersection_point_t[i], 4).tolist()] + [n_star_t[i]] for i in range(len(n_star_t))] if n_star_t != ['N/A'] else [['THEO'] + ['N/A'] + ['N/A'] ]
    data_emp = [['EMPI'] + [np.round(intersection_point_e[i], 4).tolist()] + [n_star_e[i]] for i in range(len(n_star_e))] if n_star_e != ['N/A'] else [['EMPI'] + ['N/A'] + ['N/A'] ]

    data = data_theo + data_emp

    ## make table and print
    table = tabulate(data, tablefmt='grid', headers=['Loss', 'intersection point', 'N_star'])
    print('\n','N* between ' + str(dims_to_compare[0]) + 'D and ' + str(dims_to_compare[1]) + 'D classifiers: ')
    print(table)

  def print_tags_and_tables(self, dims_to_compare=None):

    progress_stream = '----------------------------------------------------------------------------------------------------'
    n_index = len(self.sim.model.N)
    N_size = len(self.sim.model.N)
    p = int(n_index * 100 / N_size)
    dims_to_compare = dims_to_compare if dims_to_compare else self.sim.dims[-2:]
    _, _, loss_bayes, d, _, _, _ = self.compile_compare(dims=dims_to_compare)

    print(' progress: ', end='')
    print(progress_stream[0:p], end='')
    print("\033[91m {}\033[00m".format(progress_stream[p:-1]) + ' ' + str(n_index) + '/' + str(N_size))
    print('n: ' + str(self.sim.model.N[-1]))
    print('N = ' + str(self.sim.model.N))
    print('Model: ', self.model_tag)
    print('Simulator: ', self.sim_tag)
    print('d: ', { key : round(self.d[key],4) for key in self.d.keys() })
    print('bayes error rate: ', { key : round(self.loss_bayes[key],4) for key in self.loss_bayes.keys() })
    self.print_N_star(dims_to_compare)
    self.print_loss()

  def show_plots(self):
    datapoints_fig = self.sim.model.data_points_plot
    loss_fig = self.plot_with_intersection()

    if datapoints_fig:
      plt.figure(figsize=(14, 4))
      fm = plt.get_current_fig_manager()
      fm.canvas.figure = datapoints_fig
      datapoints_fig.canvas = fm.canvas
      plt.figure(datapoints_fig.number)

    plt.figure(loss_fig.number)
    plt.show()
    plt.close()

  def print_report(self, dims_to_compare=None):
    """Print report

    Parameters:
      dims_to_compare (list of int or tuple of int): list of dimensionalities to be compared

    Returns:
      None

    Raises:
      TypeError: if dims_to_compare is not a list of int or tuple of int;
      ValueError: if the number of compared dimensionalities is not 2;
                  if the list of dioemnsionalities to be compared is not a subset of the list of simulated dimensionalities

    """

    if dims_to_compare and not isinstance(dims_to_compare, (list, tuple)):
      print(type(dims_to_compare))
      print(dims_to_compare)
      raise TypeError('dims_to_compare must be a list or tuple of int')

    if dims_to_compare and len(dims_to_compare) != 2:
      raise ValueError('dims_to_compare must contain exactly 2 elements')

    if dims_to_compare and not set(dims_to_compare).issubset(set(self.sim.dims)):
      raise ValueError('dims_to_compare must be a subset of the list of simulated dimensionalities')


    dims_to_compare = dims_to_compare if dims_to_compare else self.sim.dims[-2:]

    ## if in notebook, show figures before printing tables
    if self.sim.is_notebook:
      clear_output()
      plt.close()

      # show plots
      self.show_plots()

      # print sim tags and tables
      self.print_tags_and_tables(dims_to_compare)

    ## if in terminal, print tables before showing figures
    else:
      cls()
      plt.close()

      # print sim tags and tables
      self.print_tags_and_tables(dims_to_compare)

      # show plots
      self.show_plots()

  def write_to_spreadsheet(self, gc, dims_to_compare = None, verbose=True):

    """Write results to a Google Spreadsheet

    :param self: object of class Report
    :type self: Report
    :param gc: gspread client object
    :type gc: GspreadClient
    :param dims_to_compare: list of dimensionalities to be compared
    :type dims_to_compare: list of int or tuple of int
    :param verbose: print output
    :type verbose: bool
    :return: None
    :rtype: None

    :raises TypeError:
      if dims_to_compare is not a list of int or tuple of int;

    :raises ValueError:
      if the number of compared dimensionalities is not 2;
      if the list of dioemnsionalities to be compared is not a subset of the list of simulated dimensionalities

    :Example:
      >>> import os
      >>> from slacgs import Model
      >>> from slacgs import Simulator
      >>> from slacgs import GspreadClient
      >>> from slacgs import doctest_next_parameter

      >>> ## run simulation for parameter
      >>> ### choose your own parameter
      >>> ### param = [1, 1, 2, 0, 0, 0]

      >>> ### get parameter from demo.dodoctest_next_parameter()
      >>> set_report_service_conf(slacgs_password, gdrive_user_email)
      >>> param, _ = doctest_next_parameter()

      >>> ## create model object
      >>> model = Model(param, N=[2**i for i in range(1,11)], max_n=1024)

      >>> ## create simulator object
      >>> slacgs = Simulator(model, iters_per_step=1, max_steps=10, first_step=5, precision=1e-4, augmentation_until_n = 1024, verbose=False)

      >>> ## run simulation
      >>> slacgs.run() # doctest: +ELLIPSIS

      >>> ## define spreadsheet title
      >>> ## spreadsheet_title = 'title of spreadsheet'
      >>> _, spreadsheet_title = doctest_next_parameter()

      >>> ## create GspreadClient object
      >>> gc = GspreadClient(spreadsheet_title)

      >>> ## write simulation results to spreadsheet
      >>> slacgs.report.write_to_spreadsheet(gc, verbose=False)

    """
    if dims_to_compare is None:
      dims_to_compare = self.sim.dims[-2:]

    if not isinstance(dims_to_compare, (list, tuple)):
      raise TypeError('dims_to_compare must be a list or tuple of int')

    if len(dims_to_compare) != 2:
      raise ValueError('dims_to_compare must be a list or tuple of length 2')

    if not set(dims_to_compare).issubset(set(self.sim.dims)):
      raise ValueError('dims_to_compare must be a subset of the list of simulated dimensionalities')

    gc.write_loss_report_to_spreadsheet(self, verbose=verbose)
    gc.write_compare_report_to_spreadsheet(self, dims_to_compare, verbose=verbose)
    if list(dims_to_compare) == [2,3]:
      tryal = 1
      while True:
        try:
          gc.update_home_report_on_spreadsheet(self, dims_to_compare, verbose=verbose)
        except googleapiclient.errors.HttpError:
          if tryal <= 3:
            if verbose:
              print('trying again...' + str(tryal) + '/3')
            tryal += 1
            continue
          else:
            if verbose:
              print('going next...')
            break
        else:
          break


import io
import os
import re
from googleapiclient.discovery import Resource
from googleapiclient.http import MediaIoBaseUpload

from .utils import report_service_conf, set_report_service_conf


class GdriveClient:
	"""perform operations on Google Drive."""

	def __init__(self, drive_service=None, spreadsheet_service=None, gdrive_account_email=None):

		"""Constructor for GdriveClient class.

		Parameters:
			drive_service (googleapiclient.discovery.Resource): Google Drive API Resource object.
			spreadsheet_service (googleapiclient.discovery.Resource): Google Sheets API Resource object.
			gdrive_account_email (str): email of the Google account to be used to share reports folder.

		Raises:
			TypeError:
				if drive_service is not a googleapiclient.discovery.Resource object;
				if sheets_service is not a googleapiclient.discovery.Resource object;
				if gdrive_account_email is not a string.

			ValueError:
				if gdrive_account_email is not a valid email address.

		"""

		if drive_service and not isinstance(drive_service, Resource):
			raise TypeError('drive_service must be a googleapiclient.discovery.Resource object.')

		if spreadsheet_service and not isinstance(spreadsheet_service, Resource):
			raise TypeError('sheets_service must be a googleapiclient.discovery.Resource object.')

		if gdrive_account_email:
			if not isinstance(gdrive_account_email, str):
				raise TypeError('gdrive_account_email must be a string.')

			if not re.match(r"[^@]+@[^@]+\.[^@]+", gdrive_account_email):
				raise ValueError('gdrive_account_email must be a valid email address.')

		else:
			if report_service_conf['user_email']:
				gdrive_account_email = report_service_conf['user_email']
			else:
				while True:
					gdrive_account_email = input(
						"Please enter your google account email address so I can share results with you\n(if you don't have a google account, create one at https://accounts.google.com/signup)\ngoogle account email:")
					if re.match(r"[^@]+@[^@]+\.[^@]+", gdrive_account_email):
						print("Valid email address!")
						break
					else:
						print("Invalid email address. Please try again.")

		## If drive_service and sheets_service are not provided, try to get them from report_service. If report_service is not running, start it.
		if not drive_service and not spreadsheet_service:
			if not report_service_conf['drive_service'] or not report_service_conf['spreadsheet_service']:
				set_report_service_conf()

			drive_service = report_service_conf['drive_service']
			spreadsheet_service = report_service_conf['spreadsheet_service']

		self.drive_service = drive_service
		self.sheets_service = spreadsheet_service
		self.gdrive_account_email = gdrive_account_email

	def get_folder_id_by_name(self, folder_name):

		"""Get the ID of a folder by its name. If no folder with the given name is found, return None.

		:param folder_name: name of the folder.
		:type folder_name: str

		:returns: ID of the folder with the given name.
		:rtype: str or None

		:raises ValueError: if folder_name is not a string.

		Parameters:
			folder_name (str): name of the folder.

		Returns:

		"""

		if not isinstance(folder_name, str):
			raise ValueError('folder_name must be a string.')

		response = self.drive_service.files().list(
			q=f"name='{folder_name}' and mimeType='application/vnd.google-apps.folder'", spaces='drive',
			fields='files(id)').execute()
		folders = response.get('files', [])

		if len(folders) > 0:
			return folders[0]['id']  # Return the ID of the first matching folder
		else:
			return None  # Return None if no folder with the given name is found

	def create_spreadsheet(self, name, verbose=True):
		"""Create a new spreadsheet with the given name.

		:param name: name of the spreadsheet.
		:type name: str

		:param verbose: if True, print a message after creating the spreadsheet.
		:type verbose: bool

		:returns: ID of the created spreadsheet.
		:rtype: str



		:raises ValueError: if name is not a string.

		"""

		if not isinstance(name, str):
			raise ValueError('name must be a string.')

		spreadsheet = {
			'properties': {
				'title': name
			}
		}

		spreadsheet = self.sheets_service.spreadsheets().create(body=spreadsheet).execute()
		if verbose:
			print(
				f"Spreadsheet with path '{self.get_spreadsheet_path_by_id(spreadsheet['spreadsheetId'])}' has been created.")

		return spreadsheet['spreadsheetId']

	def get_spreadsheet_path_by_id(self, spreadsheet_id):
		"""Get the path of a spreadsheet by its ID.

		:param spreadsheet_id: ID of the spreadsheet.
		:type spreadsheet_id: str

		:returns: path of the spreadsheet with the given ID.
		:rtype: str

		:raises ValueError: if spreadsheet_id is not a string.

		"""

		if not isinstance(spreadsheet_id, str):
			raise ValueError('spreadsheet_id must be a string.')

		spreadsheet = self.sheets_service.spreadsheets().get(spreadsheetId=spreadsheet_id).execute()
		return spreadsheet['properties']['title']

	def create_folder(self, folder_name, parent_folder_id=None, verbose=True):
		"""Create a new folder with the given name. If parent_folder_id is not None, the folder will be created inside the folder with the given ID.

		:param folder_name: name of the folder.
		:type folder_name: str

		:param parent_folder_id: ID of the parent folder.
		:type parent_folder_id: str

		:param verbose: whether to print the path of the created folder.
		:type verbose: bool

		:returns: ID of the created folder.
		:rtype: str

		:raises ValueError: if folder_name or parent_folder_id is not a string.

		"""

		if not isinstance(folder_name, str):
			raise ValueError('folder_name must be a string.')

		if parent_folder_id is not None and not isinstance(parent_folder_id, str):
			raise ValueError('parent_folder_id must be a string or None.')

		folder_metadata = {
			'name': folder_name,
			'mimeType': 'application/vnd.google-apps.folder'
		}
		if parent_folder_id:
			folder_metadata['parents'] = [parent_folder_id]

		folder = self.drive_service.files().create(body=folder_metadata, fields='id').execute()
		if verbose:
			print(f"Folder with path '{self.get_folder_path(folder['id'])}' has been created.")

		return folder['id']

	def move_file_to_folder(self, file_id, folder_id, verbose=True):
		"""Move a file to a folder.

		:param file_id: ID of the file to be moved.
		:type file_id: str

		:param folder_id: ID of the folder to which the file will be moved.
		:type folder_id: str

		:param verbose: whether to print messages about the operation status.
		:type verbose: bool

		:raises ValueError: if file_id or folder_id is not a string.

		"""
		file_old_path = self.get_spreadsheet_path_by_id(file_id)

		file = self.drive_service.files().get(fileId=file_id, fields='parents').execute()
		previous_parents = ",".join(file.get('parents'))

		self.drive_service.files().update(fileId=file_id, addParents=folder_id, removeParents=previous_parents,
		                                  fields='id, parents').execute()
		if verbose:
			print(
				f"Spreadsheet with path '{file_old_path}' has been moved to the folder with path '{self.get_folder_path(folder_id)}'.")

	def move_folder_to_another_folder(self, folder_id, new_parent_folder_id, verbose=True):
		"""Move a folder to another folder.

		:param folder_id: ID of the folder to be moved.
		:type folder_id: str

		:param new_parent_folder_id: ID of the folder to which the folder will be moved.
		:type new_parent_folder_id: str

		:param verbose: whether to print messages about the operation status.
		:type verbose: bool

		:returns: None
		:rtype: None

		:raises ValueError: if folder_id or new_parent_folder_id is not a string.

		"""

		if not isinstance(folder_id, str):
			raise ValueError('folder_id must be a string.')

		if not isinstance(new_parent_folder_id, str):
			raise ValueError('new_parent_folder_id must be a string.')

		# Retrieve the current parents of the folder
		folder = self.drive_service.files().get(fileId=folder_id, fields='parents').execute()
		previous_parents = ",".join(folder.get('parents'))

		# Move the folder to the new parent folder
		folder = self.drive_service.files().update(fileId=folder_id, addParents=new_parent_folder_id,
		                                           removeParents=previous_parents, fields='id, parents').execute()
		if verbose:
			print(f"Folder with name '{folder.get('name')}' has been moved to the folder with name "
			      f"'{self.drive_service.files().get(fileId=new_parent_folder_id, fields='name').execute().get('name')}'.")

	def check_spreadsheet_existence(self, name):
		"""Check if a spreadsheet with the given name exists.

		:param name: name of the spreadsheet.
		:type name: str

		:returns: True if a spreadsheet with the given name exists, False otherwise.
		:rtype: bool

		:raises ValueError: if name is not a string.

		"""

		if not isinstance(name, str):
			raise ValueError('name must be a string.')

		response = self.drive_service.files().list(
			q=f"name='{name}' and mimeType='application/vnd.google-apps.spreadsheet'", spaces='drive',
			fields='files(id)').execute()
		spreadsheets = response.get('files', [])

		if len(spreadsheets) > 0:
			return True  # A spreadsheet with the specified name exists
		else:
			return False  # No spreadsheet with the specified name exists

	def get_spreadsheet_id_by_name(self, name):
		"""Get the ID of a spreadsheet with the given name.

		:param name: name of the spreadsheet.
		:type name: str

		:returns: ID of the spreadsheet with the given name.
		:rtype: str

		:raises ValueError: if name is not a string.

		"""

		if not isinstance(name, str):
			raise ValueError('name must be a string.')

		response = self.drive_service.files().list(
			q=f"name='{name}' and mimeType='application/vnd.google-apps.spreadsheet'",
			spaces='drive',
			fields='files(id)').execute()
		spreadsheets = response.get('files', [])
		if len(spreadsheets) > 0:
			return spreadsheets[0]['id']  # Return the ID of the first matching spreadsheet
		else:
			return None  # No spreadsheet with the specified name exists

	def delete_spreadsheet(self, spreadsheet_id):
		"""Delete a spreadsheet.

		:param spreadsheet_id: ID of the spreadsheet to be deleted.
		:type spreadsheet_id: str

		:raises ValueError: if spreadsheet_id is not a string.

		"""

		if not isinstance(spreadsheet_id, str):
			raise ValueError('spreadsheet_id must be a string.')

		path = self.get_spreadsheet_path_by_id(spreadsheet_id)

		self.drive_service.files().delete(fileId=spreadsheet_id).execute()
		print(f"Spreadsheet with path '{path}' has been deleted.")

	def folder_exists(self, report_folder_name):
		"""Check if a folder with the given name exists.

		:param report_folder_name: name of the folder.
		:type report_folder_name: str

		:returns: True if a folder with the given name exists, False otherwise.
		:rtype: bool

		:raises ValueError: if report_folder_name is not a string.

		"""

		if not isinstance(report_folder_name, str):
			raise ValueError('report_folder_name must be a string.')

		response = self.drive_service.files().list(
			q=f"name='{report_folder_name}' and mimeType='application/vnd.google-apps.folder'", spaces='drive',
			fields='files(id)').execute()
		folders = response.get('files', [])

		if len(folders) > 0:
			return True

	def get_root_folder_id(self):
		"""Get the ID of the root folder.

		:returns: ID of the root folder.
		:rtype: str

		"""

		response = self.drive_service.files().get(fileId='root', fields='id').execute()
		return response.get('id')

	def get_folder_path(self, folder_id):
		"""Get the path of a folder.

		:param folder_id: ID of the folder.
		:type folder_id: str

		:returns: path of the folder.
		:rtype: str

		:raises ValueError: if folder_id is not a string.

		"""

		if not isinstance(folder_id, str):
			raise ValueError('folder_id must be a string.')

		folder_path = []
		while folder_id != 'root':
			folder = self.drive_service.files().get(fileId=folder_id, fields='id, name, parents').execute()
			folder_name = folder.get('name')
			folder_path.insert(0, folder_name)
			parents = folder.get('parents')
			if parents:
				folder_id = parents[0]
			else:
				break
		return '/'.join(folder_path)

	def share_folder_with_gdrive_account(self, folder_id, verbose=True):
		"""Share a folder with the GDrive account.

		:param folder_id: ID of the folder.
		:type folder_id: str
		
		:param verbose: whether to print messages or not.
		:type verbose: bool
	
		:return: None
		:rtype: None

		:raises ValueError: if folder_id is not a string.
		
		"""

		if not isinstance(folder_id, str):
			raise ValueError('folder_id must be a string.')

		permission = {
			'type': 'user',
			'role': 'writer',
			'emailAddress': self.gdrive_account_email
		}

		self.drive_service.permissions().create(fileId=folder_id, body=permission).execute()

		if verbose:
			print(f"Folder with path '{self.get_folder_path(folder_id)}' has been shared with the GDrive account with "
			      f"email address '{self.gdrive_account_email}'.")
			print(f"link: https://drive.google.com/drive/folders/{folder_id}")

	def delete_folder(self, folder_id):
		"""Delete a folder.

		:param folder_id: ID of the folder.
		:type folder_id: str

		:raises ValueError: if folder_id is not a string.

		"""
		path = self.get_folder_path(folder_id)
		if not isinstance(folder_id, str):
			raise ValueError('folder_id must be a string.')

		self.drive_service.files().delete(fileId=folder_id).execute()
		print(f"Folder with path '{path}' has been deleted.")

	def delete_file(self, file_id):
		"""Delete a file.

		:param file_id: ID of the file.
		:type file_id: str

		"""

		self.drive_service.files().delete(fileId=file_id).execute()
		print(f"File with ID '{file_id}' has been deleted.")

	def get_file_id_by_path(self, file_path):
		"""Get the ID of a file.

		:param file_path: path of the file.
		:type file_path: str

		:returns: ID of the file.
		:rtype: str

		:raises ValueError: if file_path is not a string.

		"""

		if not isinstance(file_path, str):
			raise ValueError('file_path must be a string.')

		file_path = file_path.split('/')
		file_path = [file for file in file_path if file != '']

		file_id = 'root'
		for file_name in file_path:
			response = self.drive_service.files().list(
				q=f"name='{file_name}' and mimeType!='application/vnd.google-apps.folder' and '{file_id}' in parents",
				spaces='drive', fields='files(id)').execute()
			files = response.get('files', [])
			if len(files) > 0:
				file_id = files[0]['id']
			else:
				return None
		return file_id

	##make a function that check if folder with given path exists and return True if it does and False if it doesn't
	def folder_exists_by_path(self, folder_path):
		"""Check if a folder with the given path exists.

		Parameters:
			folder_path (str): The path to the folder you want to check.

		Returns:
			True if the folder exists, False otherwise.

		Raises:
			ValueError: If folder_path is not a string.


		"""

		if not isinstance(folder_path, str):
			raise ValueError('folder_path must be a string.')

		folder_path = folder_path.split('/')
		folder_path = [folder for folder in folder_path if folder != '']

		folder_id = 'root'
		for folder_name in folder_path:
			response = self.drive_service.files().list(
				q=f"name='{folder_name}' and mimeType='application/vnd.google-apps.folder' and parents='{folder_id}'",
				spaces='drive', fields='files(id)').execute()
			folders = response.get('files', [])

			if len(folders) > 0:
				folder_id = folders[0].get('id')
			else:
				return False
		return True

	##make a function to get folder id by path
	def get_folder_id_by_path(self, folder_path):
		"""Get the ID of a folder by its path.

		Parameters:
			folder_path (str): The path to the folder you want to get the ID of.

		Returns:
			The ID of the folder.

		Raises:
			ValueError: If folder_path is not a string.

		"""

		if not isinstance(folder_path, str):
			raise ValueError('folder_path must be a string.')

		folder_path = folder_path.split('/')
		folder_path = [folder for folder in folder_path if folder != '']

		folder_id = 'root'
		for folder_name in folder_path:
			response = self.drive_service.files().list(
				q=f"name='{folder_name}' and mimeType='application/vnd.google-apps.folder' and parents='{folder_id}'",
				spaces='drive', fields='files(id)').execute()
			folders = response.get('files', [])

			if len(folders) > 0:
				folder_id = folders[0].get('id')
			else:
				return None
		return folder_id

	##make a function to check if file with given path exists and return True if it does and False if it doesn't
	def file_exists_by_path(self, file_path):
		"""Check if a file with the given path exists.

		Parameters:
			file_path (str): The path to the file you want to check.

		Returns:
			True if the file exists, False otherwise.

		Raises:
			ValueError: If file_path is not a string.

		"""

		if not isinstance(file_path, str):
			raise ValueError('file_path must be a string.')

		file_name = os.path.basename(file_path)
		folder_path = os.path.dirname(file_path)

		folder_id = self.get_folder_id_by_path(folder_path)
		if folder_id is None:
			return False

		response = self.drive_service.files().list(
			q=f"name='{file_name}' and mimeType!='application/vnd.google-apps.folder' and parents='{folder_id}'",
			spaces='drive', fields='files(id)').execute()
		files = response.get('files', [])

		if len(files) > 0:
			return True
		else:
			return False

	def file_exists_in_folder(self, file_name, folder_id):
		"""Check if a file with the given name exists in the given folder.

		Parameters:
			file_name (str): The name of the file you want to check.
			folder_id (str): The ID of the folder you want to check.

		Returns:
			True if the file exists, False otherwise.

		Raises:
			ValueError: If file_name or folder_id is not a string.

		"""

		if not isinstance(file_name, str):
			raise ValueError('file_name must be a string.')
		if not isinstance(folder_id, str):
			raise ValueError('folder_id must be a string.')

		response = self.drive_service.files().list(
			q=f"name='{file_name}' and mimeType!='application/vnd.google-apps.folder' and parents='{folder_id}'",
			spaces='drive', fields='files(id)').execute()
		files = response.get('files', [])

		if len(files) > 0:
			return True
		else:
			return False

	def upload_file_to_drive(self, file_path, folder_id, verbose=True):
		"""
		Uploads a file to a Google Drive folder.

		Args:
				file_path (str): The path to the file you want to upload.
				folder_id (str): The ID of the Google Drive folder where you want to upload the file.

		"""

		# Prepare the file metadata
		file_name = os.path.basename(file_path)
		file_metadata = {
			'name': file_name,
			'parents': [folder_id]
		}

		# Upload the file
		media = MediaIoBaseUpload(io.BytesIO(open(file_path, 'rb').read()), mimetype='application/octet-stream',
		                          resumable=True)

		try:
			if self.file_exists_in_folder(file_name, folder_id):
				file_id = self.get_file_id_by_path(file_path)
			else:
				uploaded_file = self.drive_service.files().create(body=file_metadata, media_body=media, fields='id').execute()
				file_id = uploaded_file.get('id')

			if verbose:
				print(f"File with path '{file_path}' has been uploaded to folder: "
				      f"https://drive.google.com/drive/folders/{folder_id}.")
				print('link to file: https://drive.google.com/file/d/' + file_id)
		except Exception as e:
			print(
				'failed to upload file with path: ' + file_path + ' to folder: https://drive.google.com/drive/folders/' + folder_id)



from enum import Enum


class LossType(Enum):
  """
  Implemented Loss Functions

  - ":math:`h(\mathbf{x}) = \mathbf{w} \cdot \mathbf{x} + b`" is a linear classifier.
  - The hyperplane  :math:`h` defined by this equation represents the decision boundary that separates classes in a linearly separable space.
  - The hypothesis space :math:`H` consists of all possible linear classifiers.

  Let :math:`\hat{h}^{(D)}` be the best empirical classifier in the dictionary :math:`H` for the dataset :math:`D` with :math:`d` features and :math:`n` samples, i.e., the classifier :math:`h` in :math:`H` with the minimum empirical error rate. Under these conditions, we can define three error rates:
    - :math:`L(\hat{h}^{(D)}) =` the theoretical error rate of the classifier :math:`\hat{h}^{(D)}`
    - :math:`\hat{L}(\hat{h}^{(D)}) =` the empirical error rate of the classifier :math:`\hat{h}^{(D)}`
    - :math:`\hat{L}(\hat{h}^{(D)},D') =` the empirical error rate of the classifier :math:`\hat{h}^{(D)}` using a test dataset :math:`D'`

  The Bayes Error (also: Bayes Risk) is defined as the minimum theoretical achievable error rate:
      - :math:`\min_{h\in H} L(h) =` the smallest theoretically achievable error rate for a classifier :math:`h` in dictionary :math:`H`.

  Attributes:
    - THEORETICAL: estimated using probability theory
    - EMPIRICALTRAIN: estimated using empirical approach with train data
    - EMPIRICALTEST: estimated using empirical approach with test data

  """
  THEORETICAL = 'THEORETICAL'
  EMPIRICALTRAIN = 'EMPIRICAL_TRAIN'
  EMPIRICALTEST = 'EMPIRICAL_TEST'


class DictionaryType(Enum):
  """Implemented Dictionary Types

  Attributes:
    - LINEAR: dictionary with linear classifiers

  """

  LINEAR = 'LINEAR'



from enum import Enum
import googleapiclient
import numpy as np
import pygsheets
import math

from pygsheets.client import Client
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from .report import Report


class GspreadClient:
	"""GspreadClient performs operations using gspread, a Python API for Google Sheets. It is used to write simulation results to a spreadsheet."""

	def __init__(self, pygsheets_service, spreadsheet_title, verbose=True):
		"""Constructor for GspreadClient class.

		Parameters:
			pygsheets_service (Client): pygsheets service
			spreadsheet_title (str): title of the spreadsheet
			verbose (bool, optional): print messages to console, defaults to True

		Raises:
		  TypeError:
		    if pygsheets_service is not a pygsheets.service.client.Client object
		    if spredsheet_title is not a string
		    if verbose is not a boolean

		  ValueError:
		    if spredsheet_title is empty

		"""

		if not isinstance(pygsheets_service, pygsheets.client.Client):
			raise TypeError('pygsheets_service must be a pygsheets.client.Client object')

		if not isinstance(spreadsheet_title, str):
			raise TypeError('spredsheet_title must be a string')

		if spreadsheet_title == '':
			raise ValueError('spredsheet_title cannot be empty')

		if not isinstance(verbose, bool):
			raise TypeError('verbose must be a boolean')

		self.sh = pygsheets_service.open(spreadsheet_title)
		self.verbose = verbose

	def print_first_call_to_spreadsheet_writing_service(fn):
		"""
		Decorator that prints a message to console when a method of GspreadClient is called for the first time.
		Parameters:
			fn (function): function to be decorated

		Returns:
		  function: decorated function

		"""

		def wrapper(self, *args, **kwargs):
			if not hasattr(self, '_method_called'):
				self._method_called = True
				print('\nStarting to write to spreadsheet: ', self.sh.title, )
				print('(This may take a while, ignore the warning messages)')
				print('link to spreadsheet: ', self.sh.url, '\n')
			return fn(self, *args, **kwargs)

		return wrapper

	@print_first_call_to_spreadsheet_writing_service
	def write_loss_report_to_spreadsheet(self, report, verbose=True):
		"""write loss_report to spreadsheet, a report that contains all results of Loss estimations

    - Reports with results will be stored in a Google Spreadsheet
    - The Spreadsheets are stored in a Google Drive folder named 'slacgs.demo.<user_email>'	owned by slacgs' google service	account and shared with the user's Google Drive account.

    Reports Exported:
      - Loss Report: Contains mainly results focused on Loss Functions evaluations for each dimensionality of the model.

    Parameters:
      report (Report): report to be written to spreadsheet
      verbose (bool): print messages to console, defaults to True

    Raises:
      TypeError:
        if report is not a Report object;
        if verbose is not a boolean


    """

		if not isinstance(report, Report):
			raise TypeError('report must be a Report object')

		if not isinstance(verbose, bool):
			raise TypeError('verbose must be a boolean')

		loss_N = report.loss_N
		iter_N = report.iter_N
		bayes_loss = report.loss_bayes
		d = report.d
		model_tag = report.model_tag
		sim_tag = report.sim_tag
		loss_types = report.sim.loss_types
		N = report.sim.model.N
		dims = report.sim.dims
		rho = report.sim.model.rho

		delta_L1, delta_L2 = report.compile_delta_L_()

		# organize report tables in 2-dim lists
		matrix_N, matrix_N_log2, matrix_loss_N, matrix_iter_N, matrix_delta_L1, matrix_delta_L2 = [], [], [], [], [], []
		matrix_delta_Ltest, matrix_delta_Ltrain, bayes_loss_matrix, d_matrix, model_tag_matrix, sim_tag_matrix = [], [], [], [], [], []

		sh = self.sh

		try:
			ws_home = sh.worksheet_by_title('home')
		except pygsheets.exceptions.WorksheetNotFound:
			matrix_N.append([None])
		else:
			matrix_N.append(['=HYPERLINK( "' + ws_home.url + '"; "🏠Home" )'])

		matrix_N.append([None])
		matrix_N.append(['N'])
		matrix_N = matrix_N + np.matrix(N).T.tolist()

		matrix_N_log2.append([None])
		matrix_N_log2.append(['dim=>>'])
		matrix_N_log2.append(['log_2(N)'])
		matrix_N_log2 = matrix_N_log2 + np.log2(np.matrix(N)).T.tolist()

		matrix_loss_N.append(['simulated loss results (P(Error)):'])
		loss_matrix_width = len(dims) * len(loss_types)
		matrix_loss_N.append([str(dims[int(i / len(loss_types))]) + ' feat' for i in range(loss_matrix_width)])
		matrix_loss_N.append([loss_types[i % len(loss_types)] for i in range(loss_matrix_width)])
		aux = [np.matrix(loss_N[i][j]).T for i in list(loss_N.keys()) for j in list(loss_N[i].keys())]
		matrix_loss_N = matrix_loss_N + np.concatenate(aux, axis=1).tolist()

		matrix_iter_N.append(['#iterations (=Datasets/N): '])
		matrix_iter_N.append([str(dims[int(i / len(loss_types))]) + ' feat' for i in range(loss_matrix_width)])
		matrix_iter_N.append([loss_types[i % len(loss_types)] for i in range(loss_matrix_width)])
		aux = [np.matrix(iter_N[i][j]).T for i in list(iter_N.keys()) for j in list(iter_N[i].keys())]
		matrix_iter_N = matrix_iter_N + np.concatenate(aux, axis=1).tolist()

		if delta_L1:
			matrix_delta_L1.append(['Stochastic error:\n ∆L_1 = L(hˆ(D)) − min(h)∈H L(h)'])
			matrix_delta_L1.append([str(dim) + 'feat' for dim in dims])
			matrix_delta_L1.append(['∆L_1' for dim in dims])
			aux = [[delta_L1[dim][i] for dim in dims] for i in range(len(N))]
			matrix_delta_L1 = matrix_delta_L1 + aux

		if delta_L2:
			matrix_delta_L2.append(['Estimation error of L(hˆ(D)):\n ∆L_2 = |L(hˆ(D)) − Lˆ(hˆ(D))|'])
			matrix_delta_L2.append([str(dim) + 'feat' for dim in dims])
			matrix_delta_L2.append(['∆L_2' for dim in dims])
			aux = [[delta_L2[dim][i] for dim in dims] for i in range(len(N))]
			matrix_delta_L2 = matrix_delta_L2 + aux

		# matrix_delta_Ltest.append(['∆L_test = E(L(hˆ(D))) − min(h)∈H L(h)'])
		# matrix_delta_Ltest.append([str(dim) + 'feat' for dim in dims])
		# aux = [[delta_Ltest[dim] for dim in dims]]
		# matrix_delta_Ltest = matrix_delta_Ltest + aux

		# matrix_delta_Ltrain.append(['∆L_train = min(h)∈H L(h) − E(Lˆ(hˆ(D)))'])
		# matrix_delta_Ltrain.append([str(dim) + 'feat' for dim in dims])
		# aux = [[delta_Ltrain[dim] for dim in dims]]
		# matrix_delta_Ltrain = matrix_delta_Ltrain + aux

		bayes_loss_matrix.append(['theoretical bayes error rate\n(or Bayes Risk):\n(min(h)∈H L(h))'])
		bayes_loss_matrix.append([str(dim) + ' feat' for dim in dims])
		bayes_loss_matrix.append(['BR_' + str(dim) + ' (min(L))' for dim in dims])
		aux = [[bayes_loss[d] for d in dims] for i in range(len(N))]
		bayes_loss_matrix = bayes_loss_matrix + aux

		d_matrix.append([
			'(d_n) as dist(P_n,P_origin), \nP_n=intersect(Line_n,Ellipsse_n)\nLine_n = Line(x_1=x_2=...=x_n)\nEllipse_n = Ellip(X_n^t . cov_n^(-1) . X_n = 1))'])
		d_matrix.append(['d_' + str(dim) for dim in dims])
		aux = [d[i] for i in dims]
		d_matrix = d_matrix + [aux]

		model_tag_matrix.append(['model parameters'])
		model_tag_matrix.append(list(model_tag.keys()))
		aux = [[model_tag[key][i] if isinstance(model_tag[key], list) and i < len(model_tag[key]) else model_tag[
			key] if i == 0 else None for key in list(model_tag.keys())] for i in range(len(model_tag['rho']))]
		model_tag_matrix = model_tag_matrix + aux

		sim_tag_matrix.append(['simulator parameters'])
		sim_tag_matrix.append(list(sim_tag.keys()))
		aux = [[sim_tag[key][i] if isinstance(sim_tag[key], list) else sim_tag[key] if i == 0 else None for key in
		        list(sim_tag.keys())] for i in range(len(sim_tag['dims']))]
		sim_tag_matrix = sim_tag_matrix + aux

		# create worksheet
		colum_pointer = 1
		colum_pointer += len(matrix_N[2])
		colum_pointer += len(matrix_N_log2[2])
		colum_pointer += len(matrix_loss_N[2])
		colum_pointer += len(matrix_iter_N[2])
		colum_pointer += len(matrix_delta_L1[3])
		colum_pointer += len(matrix_delta_L2[3]) if delta_L2 else 0
		colum_pointer += len(bayes_loss_matrix[2])
		colum_pointer += len(d_matrix[2])
		colum_pointer += len(model_tag_matrix[2])
		colum_pointer += len(sim_tag_matrix[2])

		chart_height = 18
		chart_width = 6

		parameters = ['loss'] + report.sim.model.params
		ws_title_index = 0

		if (report.sim.iters_per_step * report.sim.max_steps < 1000):
			sheet_title = '[TEST]' + str(parameters)
		else:
			sheet_title = str(parameters)

		while True:
			try:
				if ws_title_index == 0:
					sh.add_worksheet(sheet_title, rows=3 + len(N) + 3 * chart_height, cols=max(colum_pointer,
					                                                                           2 + 2 * loss_matrix_width + 3 * len(
						                                                                           dims) + max(
						                                                                           len(loss_types) * chart_width,
						                                                                           len(dims) * chart_width)))
				else:
					sh.add_worksheet(sheet_title + '[' + str(ws_title_index) + ']', rows=3 + len(N) + 3 * chart_height,
					                 cols=max(colum_pointer,
					                          2 + 2 * loss_matrix_width + 3 * len(dims) + max(len(loss_types) * chart_width,
					                                                                          len(dims) * chart_width)))
			except googleapiclient.errors.HttpError:
				ws_title_index += 1
			else:
				ws = sh.worksheet_by_title(sheet_title) if ws_title_index == 0 else sh.worksheet_by_title(
					sheet_title + '[' + str(ws_title_index) + ']')
				break

		# write matrixes on worksheet
		colum_pointer = 1
		ws.update_values((1, colum_pointer), matrix_N)
		colum_pointer += len(matrix_N[2])
		ws.update_values((1, colum_pointer), matrix_N_log2)
		colum_pointer += len(matrix_N_log2[2])
		ws.update_values((1, colum_pointer), matrix_loss_N)
		colum_pointer += len(matrix_loss_N[2])
		ws.update_values((1, colum_pointer), matrix_iter_N)
		colum_pointer += len(matrix_iter_N[2])
		ws.update_values((1, colum_pointer), matrix_delta_L1)
		colum_pointer += len(matrix_delta_L1[3])
		if delta_L2:
			ws.update_values((1, colum_pointer), matrix_delta_L2)
			colum_pointer += len(matrix_delta_L2[3])
		ws.update_values((1, colum_pointer), bayes_loss_matrix)
		colum_pointer += len(bayes_loss_matrix[2])
		ws.update_values((1, colum_pointer), d_matrix)
		colum_pointer += len(d_matrix[2])
		ws.update_values((1, colum_pointer), model_tag_matrix)
		colum_pointer += len(model_tag_matrix[2])
		ws.update_values((1, colum_pointer), sim_tag_matrix)
		colum_pointer += len(sim_tag_matrix[2])
		ws.update_value((2, colum_pointer), 'Duration (h)')
		ws.update_value((3, colum_pointer), report.duration)

		ws.adjust_column_width(1, 2, pixel_size=None)

		class ChartType(Enum):
			SCATTER = 'SCATTER'
			LINE = 'LINE'

		for i in range(len(loss_types)):
			columns = [((2, 3 + i + j * len(loss_types)), (3 + len(N), 3 + i + j * len(loss_types))) for j in
			           range(len(dims))]
			ws.add_chart(((2, 2), (3 + len(N), 2)), columns, loss_types[i] + ': P(E) vs log2(N)', chart_type=ChartType.LINE,
			             anchor_cell=(4 + len(N), 3 + i * chart_width))

		for i in range(len(dims)):
			columns = [((3, 3 + i * len(loss_types) + j), (3 + len(N), 3 + i * len(loss_types) + j)) for j in
			           range(len(loss_types))] + [((3, 3 + 2 * loss_matrix_width + i + (delta_L1 != []) * len(dims) + (
					delta_L2 != []) * len(dims)), (3 + len(N), 3 + 2 * loss_matrix_width + i + (delta_L1 != []) * len(dims) + (
					delta_L2 != []) * len(dims)))]
			ws.add_chart(((3, 2), (3 + len(N), 2)), columns, str(dims[i]) + ' feature(s)' + ': P(E) vs log2(N)',
			             chart_type=ChartType.LINE, anchor_cell=(4 + len(N) + chart_height, 3 + i * chart_width))

		if delta_L1 or delta_L2:
			for i in range(len(dims)):
				columns = [
					((3, 3 + 2 * loss_matrix_width + i + j * len(dims)), (1024, 3 + 2 * loss_matrix_width + i + j * len(dims)))
					for j in range((delta_L1 != []) + (delta_L2 != []))]
				ws.add_chart(((3, 2), (3 + len(N), 2)), columns, str(dims[i]) + ' feature(s)' + ': ∆L vs log2(N)',
				             chart_type=ChartType.LINE, anchor_cell=(4 + len(N) + 2 * chart_height, 3 + i * chart_width))

		# set chart headers
		for chart in ws.get_charts():
			spec = chart.get_json()
			spec['basicChart'].update({'headerCount': 1})
			request = {
				'updateChartSpec': {
					'chartId': chart.id, "spec": spec}
			}

			ws.client.sheet.batch_update(sh.id, request)
		if verbose:
			print('sheet is over! id: ', ws.index, ' title:', ws.title)
			print('link to Loss Report: ', ws.url)

	@print_first_call_to_spreadsheet_writing_service
	def write_compare_report_to_spreadsheet(self, report, dims=None, verbose=True):

		"""write compare_report to spreadsheet, a report that compare Loss estimantions for a pair of dimensionalities and find N*.

    - Reports with results will be stored in a Google Spreadsheet
    - The Spreadsheets are stored in a Google Drive folder named 'slacgs.demo.<user_email>'	owned by slacgs' google service	account and shared with the user's Google Drive account.

    Reports Exported:
      - Compare Resport: Contains mainly results focused on comparing the performance of the Model using 2 features and 3 features.

    Parameters:
      report (Report): Report object to be exported to spreadsheet.
      dims (list[int] or tuple[int]): list of 2 dimensionalities to compare. if dims is None, the last 2 dimensionalities simulated will be compared.
      verbose (bool): if True, print the spreadsheet url. Default is True.

    Returns:
      None

    Raises:
      TypeError:
        if report is not a Report object;
        if dims is not a list[int] or tuple[int];

      ValueError:
        if dims is not a subset of the simulated dimensionalities;
        if dims is not a list of 2 ints;

    """

		if not isinstance(report, Report):
			raise TypeError('report must be a Report object')

		if dims is None:
			dims = report.sim.model.N[-2:]

		if not isinstance(dims, (list, tuple)):
			raise TypeError('dims must be a list or tuple of 2 ints')

		if len(dims) != 2:
			raise ValueError('dims must be a list or tuple of 2 ints')

		if not all(isinstance(dim, int) for dim in dims):
			raise TypeError('dims must be a list or tuple of 2 ints')

		if not all(dim in report.sim.dims for dim in dims):
			raise ValueError('dims must be a subset of the simulated dimensionalities')

		parameters = ['compare' + str(dims[0]) + '&' + str(dims[1])] + report.sim.model.params

		sh = self.sh
		ws_title_index = 0
		sim = report.sim

		## define sheet title
		if report.sim.iters_per_step * report.sim.max_steps < 1000:
			sheet_title = '[TEST]' + str(parameters)
		else:
			sheet_title = str(parameters)

		chart_height = 18
		## create worksheet
		while True:
			try:
				if ws_title_index == 0:
					sh.add_worksheet(sheet_title, rows=3 + len(report.sim.model.N) + 3 * chart_height, cols=45)
				else:
					sh.add_worksheet(sheet_title + '[' + str(ws_title_index) + ']', rows=(len(sim.model.N) + 60), cols=45)
			except googleapiclient.errors.HttpError:
				ws_title_index += 1
			else:
				ws = sh.worksheet_by_title(sheet_title) if ws_title_index == 0 else sh.worksheet_by_title(
					sheet_title + '[' + str(ws_title_index) + ']')
				break

		## compile report
		loss_N, iter_N, bayes_loss, d, intersection_points, model_tag, sim_tag = report.compile_compare(dims)
		N = report.sim.model.N
		loss_types = report.sim.loss_types
		matrix_N, matrix_N_log2, matrix_loss_N, matrix_iter_N, bayes_loss_matrix, d_matrix, intersection_points_matrix, model_tag_matrix, sim_tag_matrix = [], [], [], [], [], [], [], [], []
		colum_pointer = 0

		## link to home
		try:
			ws_home = sh.worksheet_by_title('home')
		except pygsheets.exceptions.WorksheetNotFound:
			matrix_N.append([None])
		else:
			matrix_N.append(['=HYPERLINK( "' + ws_home.url + '"; "🏠Home" )'])

		matrix_N.append([None])
		matrix_N.append(['N'])
		matrix_N = matrix_N + np.matrix(report.sim.model.N).T.tolist()
		colum_pointer += 1
		ws.update_values((1, colum_pointer), matrix_N)

		matrix_N_log2.append([None])
		matrix_N_log2.append(['dim=>>'])
		matrix_N_log2.append(['log_2(N)'])
		matrix_N_log2 = matrix_N_log2 + np.log2(np.matrix(report.sim.model.N)).T.tolist()
		colum_pointer += len(matrix_N[3])
		ws.update_values((1, colum_pointer), matrix_N_log2)

		matrix_loss_N.append(['simulated loss results (P(Error)):'])
		matrix_loss_N.append([str(dims[int(i / len(loss_types))]) + ' feat' for i in range(2 * len(loss_types))])
		matrix_loss_N.append(
			[list(loss_N[i].keys())[j] for i in list(loss_N.keys()) for j in range(len(list(loss_N[i].keys())))])
		aux = [np.matrix(loss_N[i][j]).T for i in list(loss_N.keys()) for j in list(loss_N[i].keys())]
		matrix_loss_N = matrix_loss_N + np.concatenate(aux, axis=1).tolist()
		colum_pointer += len(matrix_N_log2[2])
		ws.update_values((1, colum_pointer), matrix_loss_N)

		matrix_iter_N.append(['#iterations (=Datasets/N): '])
		matrix_iter_N.append([str(dims[int(i / len(loss_types))]) + ' feat' for i in range(2 * len(loss_types))])
		matrix_iter_N.append(
			[list(iter_N[i].keys())[j] for i in list(iter_N.keys()) for j in range(len(list(iter_N[i].keys())))])
		aux = [np.matrix(iter_N[i][j]).T for i in list(iter_N.keys()) for j in list(iter_N[i].keys())]
		matrix_iter_N = matrix_iter_N + np.concatenate(aux, axis=1).tolist()
		colum_pointer += len(matrix_loss_N[2])
		ws.update_values((1, colum_pointer), matrix_iter_N)

		bayes_loss_matrix.append(['theoretical bayes error rate (min(h)∈H L(h)) : '])
		bayes_loss_matrix.append(
			['BR_' + str(dim) + ' (min(L))' if isinstance(dim, int) else dim for dim in list(bayes_loss.keys())])
		aux = [bayes_loss[i] for i in list(bayes_loss.keys())]
		bayes_loss_matrix = bayes_loss_matrix + [aux]
		colum_pointer += len(matrix_iter_N[2])
		ws.update_values((1, colum_pointer), bayes_loss_matrix)

		d_matrix.append([
			'(d_n) as dist(P_n,P_origin), \nP_n=intersect(Line_n,Ellipsse_n)\nLine_n = Line(x_1=x_2=...=x_n)\nEllipse_n = Ellip(X_n^t . cov_n^(-1) . X_n = 1))'])
		d_matrix.append(['d_' + str(dim) if isinstance(dim, int) else dim for dim in list(d.keys())])
		aux = [d[i] for i in list(d.keys())]
		d_matrix = d_matrix + [aux]
		colum_pointer += len(bayes_loss_matrix[2])
		ws.update_values((1, colum_pointer), d_matrix)

		intersection_points_matrix.append(['intersect points between P(E) curves'])
		intersection_points_matrix.append(
			[list(intersection_points.keys())[int(i / 3)] for i in range(3 * len(list(intersection_points.keys())))])
		intersection_points_matrix.append(
			len(list(intersection_points.keys())) * list(intersection_points[list(intersection_points.keys())[0]].keys()))
		aux = []
		max_len = max([len(intersection_points[loss_type]['log_2(N*)']) for loss_type in intersection_points.keys()])
		for i in range(max_len):
			for loss_type in list(intersection_points.keys()):
				for key in list(intersection_points[loss_type].keys()):
					aux = aux + [intersection_points[loss_type][key][i]] if i < len(
						intersection_points[loss_type][key]) else aux + [None]
			intersection_points_matrix.append(aux)
			aux = []
		colum_pointer += len(d_matrix[2])
		ws.update_values((1, colum_pointer), intersection_points_matrix)

		model_tag_matrix.append(['model parameters'])
		model_tag_matrix.append(list(model_tag.keys()))
		aux = [[model_tag[key][i] if isinstance(model_tag[key], list) and i < len(model_tag[key]) else model_tag[
			key] if i == 0 else None for key in list(model_tag.keys())] for i in range(len(model_tag['rho']))]
		model_tag_matrix = model_tag_matrix + aux
		colum_pointer += len(intersection_points_matrix[2])
		ws.update_values((1, colum_pointer), model_tag_matrix)

		sim_tag_matrix.append(['simulator parameters'])
		sim_tag_matrix.append(list(sim_tag.keys()))
		aux = [[sim_tag[key][i] if isinstance(sim_tag[key], list) else sim_tag[key] if i == 0 else None for key in
		        list(sim_tag.keys())] for i in range(len(sim_tag['dims']))]
		sim_tag_matrix = sim_tag_matrix + aux
		colum_pointer += len(model_tag_matrix[2])
		ws.update_values((1, colum_pointer), sim_tag_matrix)

		colum_pointer += len(sim_tag_matrix[2])
		ws.update_value((2, colum_pointer), 'Duration (h)')
		ws.update_value((3, colum_pointer), report.duration)

		class ChartType(Enum):
			SCATTER = 'SCATTER'
			LINE = 'LINE'

		ws.adjust_column_width(1, 2, pixel_size=None)
		ws.update_value((1, 8), '---->\nDATA\nIN\nHIDDEN\nCELLS\n---->')
		ws.hide_dimensions(9, 14, dimension='COLUMNS')

		# add charts
		for i in range(len(loss_types)):
			columns = [((2, 3 + i), (2 + len(N), 3 + i)),
			           ((2, 3 + i + len(loss_types)), (2 + len(N), 3 + i + len(loss_types)))]
			ws.add_chart(((2, 2), (2 + len(N), 2)), columns, loss_types[i] + ': P(E) vs log2(N)', chart_type=ChartType.LINE,
			             anchor_cell=(4 + len(N) + chart_height * i, 3))

		# set chart headers
		for chart in ws.get_charts():
			spec = chart.get_json()
			spec['basicChart'].update({'headerCount': 1})
			request = {
				'updateChartSpec': {
					'chartId': chart.id, "spec": spec}
			}

			ws.client.sheet.batch_update(sh.id, request)
		if verbose:
			print('sheet is over! id: ', ws.index, ' title:', ws.title)
			print('link to Compare Report: ', ws.url)

	@print_first_call_to_spreadsheet_writing_service
	def update_home_report_on_spreadsheet(self, report, dims=None, verbose=True):
		"""update home report of a spreadsheet, a report that contains a summary of N* results for a pair of dimensionalities.

    - Reports with results will be stored in a Google Spreadsheet
    - The Spreadsheets are stored in a Google Drive folder named 'slacgs.demo.<user_email>'	owned by slacgs' google service	account and shared with the user's Google Drive account.

    Reports Exported:
      - Home Report (Scenario): Contains results from all simulations in a Scenario and links to the other reports. (available only for comparison between 2D and 3D)

    Parameters:
      report (Report): Report containnig simulation results
      dims (list[int] or tuple[int]): pair of dimensionalyties to be compared, if None, report.sim.dims[-2:] will be used
      verbose (bool): if True, prints the spreadsheet url

    Raises:
      TypeError:
        if report is not a Report object
        if dims is not a list[int] or tuple[int]
        if verbose is not a bool

      ValueError:
        if dims length is not 2
        if dims is not a subset of report.sim.dims

    Warning:
			this report is available only for 3D models.

    """
		# create worksheet
		chart_height = 18
		chart_width = 8

		sh = self.sh
		sheet_title = 'home'
		new_home = False

		try:
			sh.add_worksheet(sheet_title, rows=1000, cols=90)
		except googleapiclient.errors.HttpError:
			pass
		else:
			new_home = True
		finally:
			ws = sh.worksheet_by_title(sheet_title)
			ws.index = 0

		sigma, rho, loss_bayes, d, log2_N_star_dict, loss_N_0, loss_N_1, time_spent_gen, time_spent_loss_type, time_spent_n, time_spent_dim, time_ratio_gen, time_ratio_loss_type, time_ratio_n, time_ratio_dim, iter_ratio_per_loss_type, iter_ratio_per_n, iter_ratio_per_dim = report.compile_N(
			dims)
		loss_types = report.sim.loss_types
		loss_types_n_star = list(log2_N_star_dict.keys())
		dims_sim = report.sim.dims

		parameters_1 = ['compare' + str(dims[0]) + '&' + str(dims[1])] + report.sim.model.params
		parameters_2 = ['loss'] + report.sim.model.params
		sheet_title_aux_1 = sheet_title_aux_2 = ''

		table_params = ws.get_values((2, 1), (ws.rows, len(report.sim.model.params)), value_render='FORMULA')

		url_compare = url_loss = []
		for ws_aux_index in range(6):
			if (report.sim.iters_per_step * report.sim.max_steps < 1000):
				sheet_title_aux_1 = '[TEST]' + str(parameters_1) if not ws_aux_index else '[TEST]' + str(
					parameters_1) + '[' + str(ws_aux_index) + ']'
				sheet_title_aux_2 = '[TEST]' + str(parameters_2) if not ws_aux_index else '[TEST]' + str(
					parameters_2) + '[' + str(ws_aux_index) + ']'
			else:
				sheet_title_aux_1 = str(parameters_1) if not ws_aux_index else str(parameters_1) + '[' + str(ws_aux_index) + ']'
				sheet_title_aux_2 = str(parameters_2) if not ws_aux_index else str(parameters_2) + '[' + str(ws_aux_index) + ']'
			try:
				ws_aux_1 = sh.worksheet_by_title(sheet_title_aux_1)
				ws_aux_2 = sh.worksheet_by_title(sheet_title_aux_2)
			except pygsheets.exceptions.WorksheetNotFound:
				continue
			else:
				url_compare = ws_aux_1.url
				url_loss = ws_aux_2.url
				break

		if not url_compare or not url_loss:
			url_compare = '<link not found>'
			url_loss = '<link not found>'

		indicators = [loss_bayes['diff'], loss_bayes['ratio'], d['diff'], d['ratio'], d[dims[0]], loss_bayes[dims[0]],
		              d[dims[1]], loss_bayes[dims[1]]]

		links = ['=HYPERLINK( "' + url_compare + '"; "' + sheet_title_aux_1 + '" )',
		         '=HYPERLINK( "' + url_loss + '"; "' + sheet_title_aux_2 + '" )']

		N_count_reported = 10

		cenario = 0
		rho_matrix = report.sim.model.rho_matrix
		dim = report.sim.model.dim
		param = ''
		params = report.sim.model.params

		if dim == 3:
			if rho_matrix[0][1] == 0 and rho_matrix[0][2] == 0 and rho_matrix[1][2] == 0:
				cenario = 1
				param_index = 2
				param = 'σ_3'
			elif sigma[0] == sigma[1] and rho_matrix[0][2] == 0 and rho_matrix[1][2] == 0 and rho_matrix[0][1] != 0:
				cenario = 2
				param_index = 3
				param = 'ρ_12'
			elif rho_matrix[0][1] == 0 and rho_matrix[0][2] == rho_matrix[1][2] and sigma[0] == sigma[1]:
				cenario = 3
				param_index = 4
				param = 'ρ_13=ρ_23'
			elif sigma[0] == sigma[1] and rho_matrix[0][2] == rho_matrix[1][2] and abs(rho_matrix[0][2]) <= math.sqrt(
					(1 + rho_matrix[0][1]) / 2):
				cenario = 3
				param_index = 4
				param = 'ρ_13=ρ_23'

		loss_N_1 = [['L|' + str(dims[1]) + 'feat| ' + param + '=' + str(params[param_index])] + loss_N_1[
		                                                                                        i * (N_count_reported + 1):(
				                                                                                                                   i + 1) * (
				                                                                                                                   N_count_reported + 1)]
		            for i in range(len(loss_types))]
		loss_N_1 = [loss_N_1[i][j] for i in range(len(loss_N_1)) for j in range(len(loss_N_1[i]))]

		time_spent_n = [param + '=' + str(params[param_index])] + time_spent_n
		time_spent_dim = [param + '=' + str(params[param_index])] + time_spent_dim
		time_ratio_n = [param + '=' + str(params[param_index])] + time_ratio_n
		time_ratio_dim = [param + '=' + str(params[param_index])] + time_ratio_dim
		iter_ratio_per_n = [param + '=' + str(params[param_index])] + iter_ratio_per_n
		iter_ratio_per_dim = [param + '=' + str(params[param_index])] + iter_ratio_per_dim

		cost = time_spent_gen + time_spent_loss_type + time_spent_n + time_spent_dim + time_ratio_gen + time_ratio_loss_type + time_ratio_n + time_ratio_dim + iter_ratio_per_loss_type + iter_ratio_per_n + iter_ratio_per_dim

		report_row = sigma + rho + indicators + list(log2_N_star_dict.values()) + loss_N_1 + cost + links
		table, title = [], []
		param_values_min = param_values_max = -1

		if not new_home:
			title = ws.get_values((1, 1), (3, ws.cols), value_render='FORMULA')
			table = ws.get_values((4, 1), (ws.rows, ws.cols), value_render='FORMULA')

			param_value = str(params[param_index]) if cenario == 2 else '0'
			loss_N_0 = [['L|' + str(dims[0]) + 'feat| ' + param + '=' + param_value] + loss_N_0[
			                                                                           i * (N_count_reported + 1):(i + 1) * (
					                                                                           N_count_reported + 1)] for i in
			            range(len(loss_types))]
			loss_N_0 = [loss_N_0[i][j] for i in range(len(loss_N_0)) for j in range(len(loss_N_0[i]))]
			report_row_minor_dim = sigma[:2] + [''] + rho[:1] + [''] * 2 + [''] * len(indicators) + [''] * len(
				loss_types_n_star) + loss_N_0 + [''] * 2
			table.append(report_row_minor_dim)

			table.append(report_row)

			sort_loss_column_index_1 = len(sigma) + len(rho) + len(indicators) - 1

			param_values_min = 2 ** 20
			param_values_max = -2 ** 20
			for i in range(len(table)):
				if table[i][2] == None or table[i][2] == '':
					table[i][2] = 0
					table[i][4] = 0
					table[i][5] = 0
					table[i][sort_loss_column_index_1] = 0
					if table[i][0] == '':
						table[i][0] = table[i - 1][0]
						table[i][1] = table[i - 1][1]
						table[i][2] += 0.1
						table[i][3] = table[i - 1][3]
				else:
					param_values_min = table[i][param_index] if table[i][param_index] < param_values_min else param_values_min
					param_values_max = table[i][param_index] if table[i][param_index] > param_values_max else param_values_max

			table.sort(key=lambda row: (row[0], row[1], row[2], row[3], row[4], row[5], row[sort_loss_column_index_1]))
			for i in range(len(table)):
				if table[i][2] == 0:
					table[i][2] = ''
					table[i][4] = ''
					table[i][5] = ''
					table[i][sort_loss_column_index_1] = ''
				elif table[i][2] == 0.1:
					table[i][0] = ''
					table[i][1] = ''
					table[i][2] = ''
					table[i][3] = ''
					table[i][4] = ''
					table[i][5] = ''
					table[i][sort_loss_column_index_1] = ''


		else:
			table = []
			table.append(report_row)

			sigma_title = ['σ_' + str(i + 1) for i in range(dims[-1])]
			rho_title = ['ρ_' + str(i + 1) + str(j + 1) for i in range(dims[-1]) for j in range(i + 1, dims[-1])]
			br_title = ['BR_' + str(dims[0]) + '-' + 'BR_' + str(dims[1]), 'BR_' + str(dims[0]) + '/' + 'BR_' + str(dims[1])]
			d_title = ['d_' + str(dims[0]) + '-' + 'd_' + str(dims[1]), 'd_' + str(dims[0]) + '/' + 'd_' + str(dims[1])]
			single_br_d_title = ['d_' + str(dims[0]), 'BR_' + str(dims[0]), 'd_' + str(dims[1]), 'BR_' + str(dims[1])]

			n_star_title = ['log2(N*) ' + str(loss_type) for loss_type in loss_types_n_star]
			title = sigma_title + rho_title + br_title + d_title + single_br_d_title + n_star_title

			L_title_0_aux = [[loss_type + ' Loss'] + [None] * (N_count_reported + 1) for loss_type in loss_types]
			L_title_0_aux = [L_title_0_aux[i][j] for i in range(len(L_title_0_aux)) for j in range(len(L_title_0_aux[i]))]
			L_title_0 = [''] * len(title) + L_title_0_aux
			L_title_1 = [''] * len(title) + [
				'n=' + str(2 ** (i)) if i > 0 and i < (N_count_reported + 1) else None if i == 0 else 'min(L)' for i in
				range(N_count_reported + 2)] * len(loss_types) + ['time consumption (h)', ''] + [''] * (
					            len(loss_types) + 1 + len(dims_sim) + 1 + N_count_reported) + ['% consumption'] + [''] * (
					            2 + len(loss_types) + 1 + len(dims_sim) + 1 + N_count_reported + 1) + ['% iter'] + [''] * (
					            len(loss_types) + len(dims_sim) + N_count_reported - 1)
			L_title_2 = ['L|' + str(dims[0]) + 'feat|n=' + str(2 ** (i)) if i > 0 and i < (
					N_count_reported + 1) else None if i == 0 else 'min(L)| ' + str(dims[0]) + 'feat' for i in
			             range(N_count_reported + 2)] * len(loss_types)

			cost_title = ['slacgs', 'iters'] + [loss_type for loss_type in loss_types] + [''] + ['n=' + str(2 ** (i + 1)) for
			                                                                                     i in
			                                                                                     range(N_count_reported)] + [
				             ''] + ['dim=' + str(dim) for dim in dims_sim] + ['slacgs', 'iter'] + [loss_type for loss_type in
			                                                                                     loss_types] + [''] + [
				             'n=' + str(2 ** (i + 1)) for i in range(N_count_reported)] + [''] + ['dim=' + str(dim) for dim in
			                                                                                    dims_sim] + [loss_type for
			                                                                                                 loss_type in
			                                                                                                 loss_types] + [
				             ''] + ['n=' + str(2 ** (i + 1)) for i in range(N_count_reported)] + [''] + ['dim=' + str(dim) for
			                                                                                           dim in dims_sim]
			links_title = ['url compare report', 'url loss report']

			title = title + L_title_2 + cost_title + links_title

			pre_title = [L_title_0, L_title_1]

			param_value = str(params[param_index]) if cenario == 2 else '0'
			loss_N_0 = [['L|' + str(dims[0]) + 'feat| ' + param + '=' + param_value] + loss_N_0[
			                                                                           i * (N_count_reported + 1):(i + 1) * (
					                                                                           N_count_reported + 1)] for i in
			            range(len(loss_types))]
			loss_N_0 = [loss_N_0[i][j] for i in range(len(loss_N_0)) for j in range(len(loss_N_0[i]))]
			report_row_minor_dim = sigma[:2] + [''] + rho[:1] + [''] * 2 + [''] * len(indicators) + [''] * len(
				loss_types_n_star) + loss_N_0 + [''] * 4
			title = pre_title + [title]

			title_N = [''] * (len(params) + len(indicators) + len(loss_types_n_star)) + [
				'L|' + str(dims[1]) + 'feat|n=' + str(2 ** (i)) if i > 0 and i < (
						N_count_reported + 1) else '' if i == 0 else 'min(L)| ' + str(dims[1]) + 'feat' for i in
				range(N_count_reported + 2)] * len(loss_types) + [''] * 2
			param_values_min = param_values_max = params[param_index]
			table = [report_row_minor_dim] + [title_N] + table

		table_len = len(table)
		table = title + table

		# write matrixes on worksheet
		ws.update_values((1, 1), table)
		ws.adjust_column_width(1, ws.cols, pixel_size=None)

		if ws.get_charts():
			for chart in ws.get_charts():
				chart.delete()

		class ChartType(Enum):
			SCATTER = 'SCATTER'
			LINE = 'LINE'

		params_title = str(params[:param_index] + [param] + params[param_index + 1:]) if cenario < 3 else str(
			params[:param_index] + [param[:4]] + [param[5:]])
		param_values_title = str(param_values_min) + '<=' + str(param) + '<=' + str(param_values_max)

		# log(N*) for each loss_type
		columns = [((3, len(sigma) + len(rho) + len(indicators) + 1 + i),
		            (4 + table_len, len(sigma) + len(rho) + len(indicators) + 1 + i)) for i in
		           range(len(loss_types_n_star))]

		# log(N*) vs d_i/d_(i+1)
		ws.add_chart(((3, len(sigma) + len(rho) + 4), (4 + table_len, len(sigma) + len(rho) + 4)), [columns[0]],
		             'log2(N*)|params=' + params_title + '|' + param_values_title + ' vs ' + '(d_' + str(
			             dims[0]) + '/' + 'd_' + str(dims[1]) + ')', chart_type=ChartType.SCATTER,
		             anchor_cell=(1 + len(table) + 1, 1))
		if len(columns) > 1:
			ws.add_chart(((3, len(sigma) + len(rho) + 4), (4 + table_len, len(sigma) + len(rho) + 4)), [columns[1]],
			             'log2(N*)|params=' + params_title + '|' + param_values_title + ' vs ' + '(d_' + str(
				             dims[0]) + '/' + 'd_' + str(dims[1]) + ')', chart_type=ChartType.SCATTER,
			             anchor_cell=(1 + len(table) + 1, 13))

		ws.add_chart(((3, len(sigma) + len(rho) + 1), (4 + table_len, len(sigma) + len(rho) + 1)), columns,
		             'log2(N*)|params=' + params_title + '|' + param_values_title + ' vs ' + '(BR_' + str(
			             dims[0]) + '-' + 'BR_' + str(dims[1]) + ')', chart_type=ChartType.SCATTER,
		             anchor_cell=(1 + len(table) + 1 + 2 * chart_height, 1))
		ws.add_chart(((3, len(sigma) + len(rho) + 2), (4 + table_len, len(sigma) + len(rho) + 2)), columns,
		             'log2(N*)|params=' + params_title + '|' + param_values_title + ' vs ' + '(BR_' + str(
			             dims[0]) + '/' + 'BR_' + str(dims[1]) + ')', chart_type=ChartType.SCATTER,
		             anchor_cell=(1 + len(table) + chart_height + 1, 1))
		ws.add_chart(((3, len(sigma) + len(rho) + 3), (4 + table_len, len(sigma) + len(rho) + 3)), columns,
		             'log2(N*)|params=' + params_title + '|' + param_values_title + ' vs ' + '(d_' + str(
			             dims[0]) + '-' + 'd_' + str(dims[1]) + ')', chart_type=ChartType.SCATTER,
		             anchor_cell=(1 + len(table) + 1 + 2 * chart_height, 13))
		ws.add_chart(((3, len(sigma) + len(rho) + 4), (4 + table_len, len(sigma) + len(rho) + 4)), columns,
		             'log2(N*)|params=' + params_title + '|' + param_values_title + ' vs ' + '(d_' + str(
			             dims[0]) + '/' + 'd_' + str(dims[1]) + ')', chart_type=ChartType.SCATTER,
		             anchor_cell=(1 + len(table) + chart_height + 1, 13))

		# log(N*) vs param[i]
		ws.add_chart(((3, cenario + 2), (4 + table_len, cenario + 2)), [columns[0]],
		             'log2(N*)|params=' + params_title + '|' + param_values_title + ' vs ' + param,
		             chart_type=ChartType.SCATTER, anchor_cell=(1 + len(table) + 1 + 3 * chart_height, 1))
		if len(columns) > 1:
			ws.add_chart(((3, cenario + 2), (4 + table_len, cenario + 2)), [columns[1]],
			             'log2(N*)|params=' + params_title + '|' + param_values_title + ' vs ' + param,
			             chart_type=ChartType.SCATTER, anchor_cell=(1 + len(table) + 1 + 3 * chart_height, 13))

		# d_i/d_i+1 vs BR_i/BR_i+1
		ws.add_chart(((3, len(sigma) + len(rho) + 3), (4 + table_len, len(sigma) + len(rho) + 3)),
		             [((3, len(sigma) + len(rho) + 1), (4 + table_len, len(sigma) + len(rho) + 1))],
		             '(BR_' + str(dims[0]) + '-' + 'BR_' + str(dims[1]) + ')' + ' vs ' + '(d_' + str(
			             dims[0]) + '-' + 'd_' + str(dims[1]) + ')' + ' |params=' + params_title + '|' + param_values_title,
		             chart_type=ChartType.SCATTER, anchor_cell=(1 + len(table) + 1 + 4 * chart_height, 1))
		ws.add_chart(((3, len(sigma) + len(rho) + 4), (4 + table_len, len(sigma) + len(rho) + 4)),
		             [((3, len(sigma) + len(rho) + 2), (4 + table_len, len(sigma) + len(rho) + 2))],
		             '(BR_' + str(dims[0]) + '/' + 'BR_' + str(dims[1]) + ')' + 'vs ' + '(d_' + str(
			             dims[0]) + '/' + 'd_' + str(dims[1]) + ')' + ' |params=' + params_title + '|' + param_values_title,
		             chart_type=ChartType.SCATTER, anchor_cell=(1 + len(table) + 4 * chart_height + 1, 13))

		# BR_i vs d_i
		ws.add_chart(((3, len(sigma) + len(rho) + 5), (4 + table_len, len(sigma) + len(rho) + 5)),
		             [((3, len(sigma) + len(rho) + 6), (4 + table_len, len(sigma) + len(rho) + 6))],
		             'BR_' + str(dims[0]) + ' vs ' + 'd_' + str(
			             dims[0]) + ' |params=' + params_title + '|' + param_values_title, chart_type=ChartType.SCATTER,
		             anchor_cell=(1 + len(table) + 1 + 5 * chart_height, 1))

		# BR_i+1 vs d_i+1
		ws.add_chart(((3, len(sigma) + len(rho) + 7), (4 + table_len, len(sigma) + len(rho) + 7)),
		             [((3, len(sigma) + len(rho) + 8), (4 + table_len, len(sigma) + len(rho) + 8))],
		             'BR_' + str(dims[1]) + ' vs ' + 'd_' + str(
			             dims[1]) + ' |params=' + params_title + '|' + param_values_title, chart_type=ChartType.SCATTER,
		             anchor_cell=(1 + len(table) + 1 + 5 * chart_height, 13))

		# for i in range(len(loss_types)):
		#   for j in range(N_count_reported + 1):
		#     chart_title = 'L|n=' + str(2**(j+1)) + '|Loss=' + loss_types[i] + ' vs ' + param if j<N_count_reported else 'min(L)' + ' vs ' + param
		#     x_data = ((3, 2 + cenario),(3 + table_len, 2 + cenario))

		#     y_column = (len(sigma) + len(rho) + len(indicators) + len(loss_types_n_star) + (N_count_reported+1) + (N_count_reported+1)*(i)*2 + j + 1)
		#     y_data = ((3, y_column),(3 + table_len, y_column))

		#     ws.add_chart(x_data , [y_data], chart_title , chart_type=ChartType.SCATTER, anchor_cell=(1 + len(table) + 1 + j*chart_height  , 2*chart_width + chart_width*i))

		for i in range(len(loss_types)):
			y_data = []
			x_data = ((3 + int((table_len - 1) / 2) + 2, 2 + cenario), (3 + table_len, 2 + cenario))
			for j in range(N_count_reported + 2):
				y_column = (
						len(sigma) + len(rho) + len(indicators) + len(loss_types_n_star) + (N_count_reported + 2) * (i) + j + 1)
				y_data += [((3 + int((table_len - 1) / 2) + 1, y_column), (3 + table_len, y_column))]
				y_data += [((3, y_column), (3 + int((table_len - 1) / 2), y_column))]

			chart_title = 'L|type=' + loss_types[i] + '|params=' + params_title + '|' + param_values_title + ' vs ' + param
			ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.SCATTER,
			             anchor_cell=(1 + len(table) + 1, 21 + 2 * chart_width * i))

		for i in range(len(loss_types)):
			y_data = []
			x_data = ((2, len(params) + len(indicators) + len(loss_types_n_star) + 1),
			          (2, len(params) + len(indicators) + len(loss_types_n_star) + 1 + N_count_reported))
			for j in range(table_len):
				y_data += [((3 + (j + 1),
				             len(params) + len(indicators) + len(loss_types_n_star) + 1 + (N_count_reported + 2) * i), (
					            3 + (j + 1), len(params) + len(indicators) + len(loss_types_n_star) + 1 + N_count_reported + (
							            N_count_reported + 2) * i))]
			chart_title = 'L|type=' + loss_types[i] + '|params=' + params_title + '|' + param_values_title + ' vs n'
			ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.SCATTER,
			             anchor_cell=(1 + len(table) + 1, 21 + chart_width + 2 * chart_width * i))

		# cost charts
		# time consumption x loss_types (h) vs param
		x_data = ((3, 2 + cenario), (3 + table_len, 2 + cenario))
		y_column = len(params) + len(indicators) + len(loss_types_n_star) + (N_count_reported + 2) * len(loss_types) + 1
		y_data_gen = [((3, y_column + i), (3 + table_len, y_column + i)) for i in range(2)]
		y_column += 2
		y_data = y_data_gen + [((3, y_column + i), (3 + table_len, y_column + i)) for i in range(len(loss_types))]
		chart_title = 'time consumption / loss_type (h)' + ' |params=' + params_title + '|' + param_values_title + ' vs ' + param
		ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.LINE,
		             anchor_cell=(1 + len(table) + 1, 21 + chart_width * 6))

		# time consumption x n (h) vs param
		x_data = ((3, 2 + cenario), (3 + table_len, 2 + cenario))
		y_column += len(loss_types) + 1
		y_data = [((3, y_column + i), (3 + table_len, y_column + i)) for i in range(N_count_reported)]
		chart_title = 'time consumption / n (h)' + ' |params=' + params_title + '|' + param_values_title + ' vs ' + param
		ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.LINE,
		             anchor_cell=(1 + len(table) + 1 + 1 * chart_height, 21 + chart_width * 6))

		# time consumption x dim (h) vs param
		x_data = ((3, 2 + cenario), (3 + table_len, 2 + cenario))
		y_column += N_count_reported + 1
		y_data = y_data_gen + [((3, y_column + i), (3 + table_len, y_column + i)) for i in range(len(dims_sim))]
		chart_title = 'time consumption / dim (h)' + ' |params=' + params_title + '|' + param_values_title + ' vs ' + param
		ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.LINE,
		             anchor_cell=(1 + len(table) + 1 + 2 * chart_height, 21 + chart_width * 6))

		# % consumption x loss_type (h) vs param
		x_data = ((3, 2 + cenario), (3 + table_len, 2 + cenario))
		y_column += len(dims_sim)
		y_data_gen = [((3, y_column + i), (3 + table_len, y_column + i)) for i in range(2)]
		y_column += 2
		y_data = [((3, y_column + i), (3 + table_len, y_column + i)) for i in range(len(loss_types))]
		chart_title = '% consumption / loss_type (h)' + ' |params=' + params_title + '|' + param_values_title + ' vs ' + param
		ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.LINE,
		             anchor_cell=(1 + len(table) + 1 + 3 * chart_height, 21 + chart_width * 6))

		# % consumption x n (h) vs param
		x_data = ((3, 2 + cenario), (3 + table_len, 2 + cenario))
		y_column += len(loss_types) + 1
		y_data = [((3, y_column + i), (3 + table_len, y_column + i)) for i in range(N_count_reported)]
		chart_title = '% consumption / n (h)' + ' |params=' + params_title + '|' + param_values_title + ' vs ' + param
		ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.LINE,
		             anchor_cell=(1 + len(table) + 1 + 4 * chart_height, 21 + chart_width * 6))

		# % consumption x dim (h) vs param
		x_data = ((3, 2 + cenario), (3 + table_len, 2 + cenario))
		y_column += N_count_reported + 1
		y_data = [((3, y_column + i), (3 + table_len, y_column + i)) for i in range(len(dims_sim))]
		chart_title = '% consumption / dim (h)' + ' |params=' + params_title + '|' + param_values_title + ' vs ' + param
		ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.LINE,
		             anchor_cell=(1 + len(table) + 1 + 5 * chart_height, 21 + chart_width * 6))

		# % iterations per loss_type vs param
		x_data = ((3, 2 + cenario), (3 + table_len, 2 + cenario))
		y_column += len(dims_sim)
		y_data = [((3, y_column + i), (3 + table_len, y_column + i)) for i in range(len(loss_types))]
		chart_title = '% iter / loss_type' + ' |params=' + params_title + '|' + param_values_title + ' vs ' + param
		ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.LINE,
		             anchor_cell=(1 + len(table) + 1 + 6 * chart_height, 21 + chart_width * 6))

		# % iterations per n vs param
		x_data = ((3, 2 + cenario), (3 + table_len, 2 + cenario))
		y_column += len(loss_types) + 1
		y_data = [((3, y_column + i), (3 + table_len, y_column + i)) for i in range(N_count_reported)]
		chart_title = '% iter / n (h)' + ' |params=' + params_title + '|' + param_values_title + ' vs ' + param
		ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.LINE,
		             anchor_cell=(1 + len(table) + 1 + 7 * chart_height, 21 + chart_width * 6))

		# % iterations per dim (h) vs param
		x_data = ((3, 2 + cenario), (3 + table_len, 2 + cenario))
		y_column += N_count_reported + 1
		y_data = [((3, y_column + i), (3 + table_len, y_column + i)) for i in range(len(dims_sim))]
		chart_title = '% iter / dim (h)' + ' |params=' + params_title + '|' + param_values_title + ' vs ' + param
		ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.LINE,
		             anchor_cell=(1 + len(table) + 1 + 8 * chart_height, 21 + chart_width * 6))

		# time consumption per param vs n
		y_column = len(params) + len(indicators) + len(loss_types_n_star) + (N_count_reported + 2) * len(
			loss_types) + 2 + len(loss_types) + 1
		x_data = ((3, y_column), (3, y_column + N_count_reported))
		y_data = [((4 + i, y_column), (4 + i, y_column + N_count_reported)) for i in range(table_len)]
		chart_title = 'time consumption / slacgs' + ' |params=' + params_title + '|' + param_values_title + ' vs n'
		ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.LINE,
		             anchor_cell=(1 + len(table) + 1 + 0 * chart_height, 21 + chart_width * 7))

		# % consumption per param vs n
		y_column += 1 + N_count_reported + 1 + len(dims_sim) + 2 + len(loss_types)
		x_data = ((3, y_column), (3, y_column + N_count_reported))
		y_data = [((4 + i, y_column), (4 + i, y_column + N_count_reported)) for i in range(table_len)]
		chart_title = '% consumption / slacgs' + ' |params=' + params_title + '|' + param_values_title + ' vs n'
		ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.LINE,
		             anchor_cell=(1 + len(table) + 1 + 1 * chart_height, 21 + chart_width * 7))

		# % iterations per param vs n
		y_column += 1 + N_count_reported + 1 + len(dims_sim) + len(loss_types)
		x_data = ((3, y_column), (3, y_column + N_count_reported))
		y_data = [((4 + i, y_column), (4 + i, y_column + N_count_reported)) for i in range(table_len)]
		chart_title = '% iter / slacgs' + ' |params=' + params_title + '|' + param_values_title + ' vs n'
		ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.LINE,
		             anchor_cell=(1 + len(table) + 1 + 2 * chart_height, 21 + chart_width * 7))

		# time consumption per param vs dim
		y_column = len(params) + len(indicators) + len(loss_types_n_star) + (N_count_reported + 2) * len(
			loss_types) + 2 + len(loss_types) + 1 + N_count_reported + 1
		x_data = ((3, y_column), (3, y_column + len(dims_sim)))
		y_data = [((4 + i, y_column), (4 + i, y_column + len(dims_sim))) for i in range(table_len)]
		chart_title = 'time consumption / slacgs' + ' |params=' + params_title + '|' + param_values_title + ' vs dim'
		ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.LINE,
		             anchor_cell=(1 + len(table) + 1 + 0 * chart_height, 21 + chart_width * 8))

		# % consumption per param vs dim
		y_column += 1 + len(dims_sim) + 2 + len(loss_types) + 1 + N_count_reported
		x_data = ((3, y_column), (3, y_column + len(dims_sim)))
		y_data = [((4 + i, y_column), (4 + i, y_column + len(dims_sim))) for i in range(table_len)]
		chart_title = '% consumption / slacgs' + ' |params=' + params_title + '|' + param_values_title + ' vs dim'
		ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.LINE,
		             anchor_cell=(1 + len(table) + 1 + 1 * chart_height, 21 + chart_width * 8))

		# % iterations per param vs dim
		y_column += 1 + len(dims_sim) + len(loss_types) + 1 + N_count_reported
		x_data = ((3, y_column), (3, y_column + len(dims_sim)))
		y_data = [((4 + i, y_column), (4 + i, y_column + len(dims_sim))) for i in range(table_len)]
		chart_title = '% iter / slacgs' + ' |params=' + params_title + '|' + param_values_title + ' vs dim'
		ws.add_chart(x_data, y_data, chart_title, chart_type=ChartType.LINE,
		             anchor_cell=(1 + len(table) + 1 + 2 * chart_height, 21 + chart_width * 8))

		charts = ws.get_charts()
		for i in range(len(charts)):
			spec = charts[i].get_json()
			spec['basicChart'].update({'headerCount': 1})
			spec['basicChart'].update({'legendPosition': 'TOP_LEGEND'})

			if (i == 1 or i == 7) and len(loss_types_n_star) == 2:
				spec['basicChart']['series'][0].update(
					{'colorStyle': {'rgbColor': {"red": 1, "green": 0, "blue": 0, "alpha": 1}}})

			request = {
				'updateChartSpec': {
					'chartId': charts[i].id, "spec": spec}}
			ws.client.sheet.batch_update(sh.id, request)

		if verbose:
			print('sheet is over! id: ', ws.index, ' title:', ws.title)
			print('link to home: ', ws.url)

	def param_not_in_home(self, params):
		""" check if params is already in home sheet

    Parameters:
      params (list[float] or tuple[float]): list of parameters containing Sigmas and Rhos

    Returns:
      bool: True if params is not in home sheet, False otherwise

    """
		sh = self.sh
		ws_home = sh.worksheet(value=0)
		already_done = ws_home.get_values((4, 1), (ws_home.rows, 6), value_render='FORMULA')

		return not (params in already_done)
